#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#define SIZE 100
#define MUTATION_CHANCE 0.05

/* This is, try permuting until you find a better one */
/* If not defined we do a full GA algorithm */
/* #define SIMPLE */

/* Only one set of these should be active */
#define MODFUNC permute
#define NEWFUNC make_new

/*#define MODFUNC mutate
#define NEWFUNC make_random*/

typedef struct {
	int score;
	int area[8][8];
} Position;

static void
print_dot (int iter)
{
	if ((iter & 0xfff) == 0) {
		printf (".");
		fflush (stdout);
	}
}

static void
permute (Position *pos)
{
	int i, m, n;

	pos->score = -1;

	m = (rand () >> 4) % 8;
	n = (rand () >> 4) % 8;

	if ((rand () >> 4) & 1) {
		for (i = 0; i < 8; i++) {
			int tmp = pos->area[i][m];
			pos->area[i][m] = pos->area[i][n];
			pos->area[i][n] = tmp;
		}
	} else {
		for (i = 0; i < 8; i++) {
			int tmp = pos->area[m][i];
			pos->area[m][i] = pos->area[n][i];
			pos->area[n][i] = tmp;
		}
	}
}


static void
make_new (Position *pos)
{
	int i;

	memset (pos, 0, sizeof (Position));

	pos->score = -1;

	for (i = 0; i < 8; i++) {
		pos->area[i][i] = 1;
	}

	for (i = 0; i < 8; i++) {
		permute (pos);
	}
}

static void
mutate (Position *pos)
{
	int i, j;

	pos->score = -1;

	for (i = 0; i < 8; i++) {
		for (j = 0; j < 8; j++) {
			if (((double)rand ()) / ((double)RAND_MAX) < MUTATION_CHANCE)
				pos->area[i][j] = ! pos->area[i][j];
		}
	}
}

static void
make_random (Position *pos)
{
	int i, j;

	pos->score = -1;

	for (i = 0; i < 8; i++) {
		for (j = 0; j < 8; j++) {
			pos->area[i][i] = ((rand () >> 4) & 1);
		}
	}
}

static int
attacks (int i, int j, int x, int y)
{
	if (i == x ||
	    j == y ||
	    (x-y) == (i-j) ||
	    (x+y) == (i+j))
		return 1;
	else
		return 0;
}

static int
count_attacks (Position *pos, int x, int y)
{
	int i, j;
	int score = 0;

	for (i = 0; i < 8; i++) {
		for (j = 0; j < 8; j++) {
			if ((i != x || j != y) &&
			    pos->area[i][j] &&
			    attacks (i, j, x, y))
				score ++;
		}
	}

	return score;
}

static int
count_score (Position *pos)
{
	int i, j;
	int score = 0;
	int queens = 0;
	int off;

	/* got a cached value */
	if (pos->score >= 0)
		return pos->score;

	for (i = 0; i < 8; i++) {
		for (j = 0; j < 8; j++) {
			if (pos->area[i][j]) {
				queens ++;
				score += count_attacks (pos, i, j);
			}
		}
	}

	/* make sure we never reach zero when there is
	 * less then 8 queens */
	off = abs (queens - 8);
	score += off * off;

	/* cache the value */
	pos->score = score;

	return score;
}

static void
print_pos (Position *pos)
{
	int i, j;

	for (i = 0; i < 8; i++) {
		for (j = 0; j < 8; j++) {
			printf ("%s", pos->area[i][j] ? " Q" : " .");
		}
		printf ("\n");
	}
}

#ifndef SIMPLE
int
main (void)
{
	int i, iter;
	Position pos[SIZE];
	Position *winpos;
	int run_avg;

	srand (getpid () * time (NULL));

	memset (pos, 0, SIZE * sizeof (Position));

	for (i = 0; i < SIZE; i++) {
		NEWFUNC (&pos[i]);
	}

	iter = 0;
	winpos = NULL;

	run_avg = 0;

	do {
		int m, n;
		int score_m, score_n;

		m = (rand () >> 4) % SIZE;
		n = (rand () >> 4) % SIZE;

		if (m == n)
			continue;

		score_m = count_score (&pos[m]);
		if (score_m == 0) {
			winpos = &pos[m];
			break;
		}
		score_n = count_score (&pos[n]);
		if (score_n == 0) {
			winpos = &pos[n];
			break;
		}

		run_avg = ( (9 * run_avg) + ((score_m + score_n) / 2) ) / 10;
		if (iter > 0 && (iter & 0xffff) == 0) {
			printf ("\nAverage score (last 10 fights): %d\n", run_avg);
		}

		if (score_n > score_m) {
			memcpy (&pos[n], &pos[m], sizeof (Position));
			MODFUNC (&pos[n]);
		} else {
			memcpy (&pos[m], &pos[n], sizeof (Position));
			MODFUNC (&pos[m]);
		}

		iter ++;
		print_dot (iter);
	} while (winpos == NULL);

	printf ("\n------------- Got it! (iterations: %d) --------------\n", iter);
	print_pos (winpos);

	return 0;
}
#else
int
main (void)
{
	int iter, best_score, score;
	Position pos;
	Position best;

	srand (getpid () * time (NULL));

	NEWFUNC (&pos);
	memcpy (&best, &pos, sizeof (Position));
	best_score = count_score (&pos);

	iter = 1;

	while (best_score != 0) {
		MODFUNC (&pos);
		score = count_score (&pos);
		if (score < best_score) {
			best_score = score;
			memcpy (&best, &pos, sizeof (Position));
			printf ("\nBest score so far: %d\n", best_score);
		}
		iter ++;
		print_dot (iter);
	}

	printf ("\n------------- Got it! (iterations: %d) --------------\n", iter);
	print_pos (&best);

	return 0;
}
#endif
